/***
 *  Functionality for arithmetic dihedral invariants.
 *
 *  Distributed under the terms of the GNU Lesser General Public License (L-GPL)
 *                  http://www.gnu.org/licenses/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  Copyright 2012, R. Lercier, J. Sijsling, C. Ritzenthaler
 */

/***
 *
 * Details about the algorithms implemented here can be found in [LeJeRi2012].
 * 
 * Bibliography
 * 
 * [LeJeRi2012] "Explicit Galois Obstruction and Descent for Hyperelliptic
 * Curves with Tamely Cyclic Reduced Automorphism Group", Preprint, 2012
 *
 *********************************************************************/

function TransformForm(T,f);
  //Input :   A homogeneous binary form  f  and an invertible matrix  T  over the same ground field.
  //Output :  The transformation of  f  by  T  .

  S2 := Parent(f); x := S2.1; z := S2.2;

  Ti := T^(-1);
  h := hom<S2 -> S2 | [Ti[1,1]*x + Ti[1,2]*z, Ti[2,1]*x + Ti[2,2]*z]>;
  
  return h(f);

end function;

function HasCyclicAutomorphismGroup(f);
 //Input :   A homogeneous binary form  f  whose geometric automorphism group is cyclic.
 //Output :  A boolean indicating whether or not the geometric automorphism group of  f  is cyclic,
 //          f  over an extension,
 //          a generator  alpha  of the geometric automorphism group over this extension,
 //          and the order of  alpha  .
 
 d := Degree(f);
 g := Integers()!((d - 2)/2);
 S := Parent(f); X := S.1; Z := S.2;
 K := BaseRing(S);
 R := PolynomialRing(K); t := R.1;
 h := hom<S -> R | [t,1]>;
 hf := h(f);

 gtest := g eq 3;
 ntest := false;
 if gtest then
     I := ShiodaInvariants(h(f) : normalize := true);
     //Reading off automorphisms from Shioda invariants:
     G := GeometricAutomorphismGroupFromShiodaInvariants(I);
     ntest := (#G eq 4);
 end if;

 //We only apply intelligent methods for the interesting case in genus 3.
 if gtest and ntest then
     test,T := IsGL2Equivalent(hf,hf,d);
     for t in T do
	 alpha := Matrix(K,2,2,t);
	 if not IsScalar(alpha) then
	     gen := alpha;
	 end if;
     end for;
     return true,f,gen,2;
 end if;
 
 //This costs way too much time:
 L := SplittingField(hf);
 S := PolynomialRing(L,2); X := S.1; Z := S.2;
 R := PolynomialRing(L); t := R.1;
 fL := S ! f;
 hfL := R ! hf;
 //Bottleneck; computing automorphisms over the splitting field of the form.
 test,T := IsGL2Equivalent(hfL,hfL,d);
 test := false;
 for t in T do
     //Determination of order:
     if not test then
	 n := 1;
	 tp := t;
	 while not IsScalar(tp) do
	     tp := t*tp;
	     n := n + 1;
	 end while;
	 if n eq #T then
	     test := true;
	     gen := t;
	 end if;
     end if;
 end for;
 return test,fL,gen,n;

end function;

function StandardForm(f,alpha);
  //Input :   A homogeneous binary form  f  and a generator  alpha  of its geometric automorphism group.
  //Output :  A standard form of  f  over an extension.

  K := BaseRing(Parent(f));
  R := PolynomialRing(K); t := R.1;
  g := R ! CharacteristicPolynomial(alpha);
  if IsIrreducible(g) then
      L := AbsoluteField(SplittingField(g));
  else
      L := K;
  end if;

  S := PolynomialRing(L,2); X := S.1; Z := S.2;
  R := PolynomialRing(L); t := R.1;

  fL := S ! f;
  alphaL := Matrix(L,2,2,[L ! alpha[1,1], L ! alpha[1,2], L ! alpha[2,1], L ! alpha[2,2]]);
  D,T := JordanForm(alphaL);

  //Transform such that the automorphism becomes diagonal.
  return TransformForm(T,fL);

end function;

function CaseDistinction(f,n);
  //Input :   A homogeneous binary form  f  in standard form, along with an integer  n
  //          which equals the order of its geometric reduced automorphism group.
  //Output :  Indices indicating the family it is in.

  S2 := Parent(f); x := S2.1; z := S2.2;
  i := 0;
  for s in [x,z] do
      if IsDivisibleBy(f,s) then
	  i := i + 1;
      end if;
  end for;

  d := Degree(f);
  m := Integers()!((d - i)/n);
  j := m mod 2;

  return [i,j,m,n];

end function;

function NormalizeCaseI1(f);
  //Input :   A homogeneous binary form  f  in standard form.
  //Output :  Its normalization divisible by  z  .

  S2 := Parent(f); x := S2.1; z := S2.2;
  test := IsDivisibleBy(f,z);
  h := hom<S2 -> S2 | [z,x]>;

  if test then
      return f;
  end if;

  return h(f);

end function;

function FundamentalCoefficients(f,C);
  //Input :   A homogeneous binary form  f  in standard form in the second family along with its case distinction.
  //Output :  A uniform representation of its coefficients.

  S2 := Parent(f); x := S2.1; z := S2.2;
  d := Degree(f);
  i0,j0,m0,n0 := Explode(C);

  as := [];
  i0red := i0;
  if i0red ne 0 then
      i0red := 1;
  end if;
  for i:=0 to m0 do
      j := i*n0 + i0red;
      Append(~as,MonomialCoefficient(f,x^(d-j)*z^(j)));
  end for;

  return as;

end function;

function GenericATI(as,C);
  //Input :   The fundamental coefficients  as  of a standard form and its case distinction  C  .
  //Output :  The generic ATI of  f  .

  i0,j0,m0,n0 := Explode(C);

  if j0 eq 0 then
      k := m0 div 2;
      J1 := [as[k+1]];
      W1 := [1];
      J2 := [];
      W2 := [];
      for i:=1 to k-1 do
	  Append(~J2,as[i]*as[2*k+2-i]);
	  Append(~W2,2);
      end for;
      Jrest := [];
      Wrest := [];
      Jrestp := [];
      Wrestp := [];
      for i:=1 to k do
	  Append(~Jrest,as[k]^i*as[k+i+1]);
	  Append(~Wrest,i+1);
	  Append(~Jrestp,as[k+2]^i*as[k-i+1]);
	  Append(~Wrestp,i+1);
      end for;
  else
      k := ((m0 + 1) div 2);
      J1 := [];
      W1 := [];
      J2 := [];
      W2 := [];
      for i:=1 to k-1 do
	  Append(~J2,as[i]*as[2*k+1-i]);
	  Append(~W2,2);
      end for;
      Jrest := [];
      Wrest := [];
      Jrestp := [];
      Wrestp := [];
      for i:=1 to k do
	  Append(~Jrest,as[k]^(2*i-1)*as[k+i]);
	  Append(~Wrest,2*i);
	  //We take the primed versions as well, to compute the ADI later
	  //and because they may be necessary when reconstructing:
	  Append(~Jrestp,as[k+1]^(2*i-1)*as[k-i+1]);
	  Append(~Wrestp,2*i);
      end for;
  end if;
  
  return [J1,J2,Jrest,Jrestp], [W1,W2,Wrest,Wrestp];
  
end function;

function GenericNormalizedATI(as,C);
  //Input :   The fundamental coefficients  as  of a standard form and its case distinction  C  .
  //Output :  The generic normalized ATI of  f  over the corresponding field of moduli.

  i0,j0,m0,n0 := Explode(C);
  J,W := GenericATI(as,C);
  J1,J2,Jrest,Jrestp := Explode(J);
  W1,W2,Wrest,Wrestp := Explode(W);

  //Projective space normalization to get the canonical representative over a base field that is as small as possible:
  P :=  WPSNormalize((W1 cat W2 cat Wrest cat Wrestp),(J1 cat J2 cat Jrest cat Jrestp));

  //Generating field of moduli:
  L := Parent(P[1]);
  K := sub<L | P>;
  if K eq Rationals() then
      K := RationalsAsNumberField();
  end if;
  //Coercing:
  J1new := [];
  J2new := [];
  Jrestnew := [];
  Jrestpnew := [];
  i := 1;
  for j in J1 do
      Append(~J1new,K ! P[i]);
      i := i + 1;
  end for;
  for j in J2 do
      Append(~J2new,K ! P[i]);
      i := i + 1;
  end for;
  for j in Jrest do
      Append(~Jrestnew,K ! P[i]);
      i := i + 1;
  end for;
  for j in Jrestp do
      Append(~Jrestpnew,K ! P[i]);
      i := i + 1;
  end for;
  
  return [J1new,J2new,Jrestnew,Jrestpnew],[W1,W2,Wrest,Wrestp],K;

end function;

function GenericADI(as,C);
  //Input :   The fundamental coefficients  as  of a standard form and its case distinction  C  .
  //Output :  The generic ADI of  f  .
  //          Not too many of these are created.

  i0,j0,m0,n0 := Explode(C);
  J,W := GenericATI(as,C);
  J1,J2,Jrest,Jrestp := Explode(J);
  W1,W2,Wrest,Wrestp := Explode(W);
  
  //Already symmetrized:
  I1 := J1;
  V1 := W1;
  I2 := J2;
  V2 := W2;

  //Now for the others. To start with, we only take the sums of the primed and the unprimed versions.
  Irest := [Jrest[1]];
  Vrest := [2];
  l0 := 0;
  stop := false;
  i := 2;
  while (not stop) and (i le #Jrest) do
      Append(~Irest,Jrest[i] + Jrestp[i]);
      Append(~Vrest,Wrest[i]);
      //  Append(~Irest,Jrest[i] * Jrestp[i]);
      //  Append(~Vrest,2*Wrest[i]);
      if Jrest[i] ne Jrestp[i] then
	  l0 := i;
	  stop := true;
      end if;
      i := i + 1;
  end while;

  //If  Jrest[i] equals Jrestp[i]  everywhere, then we have enough, because we can halve the ADI to reconstruct them.
  //If not, then we know an index where l0 where they differ, and the linear system
  //       J[i] +       J'[i]   = x
  //J'[l0]*J[i] + J[l0]*J'[i]   = y
  //is invertible, so we have enough invariants.
  //NB : This modifies the  I_{k,k,2}  by a factor of  2  !
  Icross := [];
  Vcross := [];
  if l0 ne 0 then
      for i:=2 to #Jrest do
	  Append(~Icross,Jrest[i]*Jrestp[l0] + Jrest[l0]*Jrestp[i]);
	  Append(~Vcross,Wrest[i] + Wrestp[l0]);
      end for;
  end if;

  return [I1,I2,Irest,Icross],[V1,V2,Vrest,Vcross];

end function;

function GenericNormalizedADI(as,C);
    //Input :   The normalized ADI/ATI  I / J  and their weight vectors  WI / WJ .
    //Output :  The normalized ADI/ATI  I / J  and their weight vectors  WI / WJ , as well as the corresponding fields of moduli.
    
    i0,j0,m0,n0 := Explode(C);
    J,W,K := GenericNormalizedATI(as,C);
    J1,J2,Jrest,Jrestp := Explode(J);
    W1,W2,Wrest,Wrestp := Explode(W);
    
    //Already symmetrized:
    I1 := J1;
    V1 := W1;
    I2 := J2;
    V2 := W2;

    //Now for the others. To start with, we only take the sums of the primed and the unprimed versions.
    Irest := [Jrest[1]];
    Vrest := [2];
    l0 := 0;
    stop := false;
    i := 2;
    while (not stop) and (i le #Jrest) do
	Append(~Irest,Jrest[i] + Jrestp[i]);
	Append(~Vrest,Wrest[i]);
	//  Append(~Irest,Jrest[i] * Jrestp[i]);
	//  Append(~Vrest,2*Wrest[i]);
	if Jrest[i] ne Jrestp[i] then
	    l0 := i;
	    stop := true;
	end if;
	i := i + 1;
    end while;
    
    //If  Jrest[i] equals Jrestp[i]  everywhere, then we have enough, because we can halve the ADI to reconstruct them.
    //If not, then we know an index where l0 where they differ, and the linear system
    //       J[i] +       J'[i]   = x
    //J'[l0]*J[i] + J[l0]*J'[i]   = y
    //is invertible, so we have enough invariants.
    Icross := [];
    Vcross := [];
    if l0 ne 0 then
	for i:=2 to #Jrest do
	    if i ne l0 then
		Append(~Icross,Jrest[i]*Jrestp[l0] + Jrest[l0]*Jrestp[i]);
		Append(~Vcross,Wrest[i] + Wrestp[l0]);
	    else
		Append(~Icross,Jrest[l0]*Jrestp[l0]);
		Append(~Vcross,2*Wrest[l0]);
	    end if;
	end for;
    end if;
    
    I := [I1,I2,Irest,Icross];
    V := [V1,V2,Vrest,Vcross];

    //Generating field of moduli:
    k := sub<K | &cat(I)>;
    if k eq Rationals() then
	k := RationalsAsNumberField();
    end if;
    //Coercing:
    I1new := [];
    I2new := [];
    Irestnew := [];
    Icrossnew := [];
    for i in I1 do
	Append(~I1new,k ! i);
    end for;
    for i in I2 do
	Append(~I2new,k ! i);
    end for;
    for i in Irest do
	Append(~Irestnew,k ! i);
    end for;
    for i in Icross do
	Append(~Icrossnew,k ! i);
    end for;
    
    Inew := [I1new,I2new,Irestnew,Icrossnew];
    
    return Inew,V,J,W,K,k;

end function;
