/***
 *  Descent functionality for hyperelliptic curves.
 *
 *  Distributed under the terms of the GNU Lesser General Public License (L-GPL)
 *                  http://www.gnu.org/licenses/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  Copyright 2012, R. Lercier, J. Sijsling, C. Ritzenthaler
 */

/***
 *
 * Details about the algorithms implemented here can be found in [LeJeRi2012].
 * 
 * Bibliography
 * 
 * [LeRi2012] "Explicit Galois Obstruction and Descent for Hyperelliptic
 * Curves with Tamely Cyclic Reduced Automorphism Group", Preprint, 2012
 *
 *********************************************************************/

function GenericPartialDescent(as,C,J,WJ,K);
  //Input :   The fundamental coefficients  as  of a standard form and its case distinction  C  .
  //Output :  A partial descent of  f  to a quadratic field.

  i0,j0,m0,n0 := Explode(C);
  J1,J2,Jrest,Jrestp := Explode(J);
  W1,W2,Wrest,Wrestp := Explode(WJ);

  //Calculations above in reverse:
  as0 := [];
  if j0 eq 0 then
      k := m0 div 2;
      if as[k] ne 0 then
	  as0[k] := K ! 1;
	  as0[k+1] := J1[1];
	  for i:=1 to k do
	      as0[k+i+1] := Jrest[i];
	  end for;
	  for i:=1 to k-1 do
	      as0[i] := J2[i]/as0[2*k+2-i];
	  end for;
      else
	  as0[k+2] := K ! 1;
	  as0[k+1] := J1[1];
	  for i:=1 to k do
	      as0[k-i+1] := Jrestp[i];
	  end for;
	  for i:=1 to k-1 do
	      as0[2*k+2-i] := J2[i]/as0[i];
	  end for;
      end if;
  else
      k := ((m0 + 1) div 2);
      if as[k] ne 0 then
	  as0[k]:= K ! 1;
	  for i:=1 to k do
	      as0[k+i] := Jrest[i];
	  end for;
	  for i:=1 to k-1 do
	      as0[i] := J2[i]/as0[2*k+1-i];
	  end for;
      else
	  as0[k+1] := K ! 1;
	  for i:=1 to k do
	      as0[k-i+1] := Jrestp[i];
	  end for;
	  for i:=1 to k-1 do
	      as0[2*k+1-i] := J2[i]/as0[i];
	  end for;
      end if;
  end if;

  //Form from coefficients:
  d := m0*n0;
  S := PolynomialRing(K,2); x := S.1; z := S.2;
  f := S ! 0;
  if i0 eq 0 then
      for i:=1 to #as0 do
	  j := (i-1)*n0;
	  f := f + as0[i]*x^(j)*z^(d-j);
      end for;
  elif i0 eq 1 then
      for i:=1 to #as0 do
	  j := (i-1)*n0;
	  f := f + as0[i]*x^(j)*z^(d-j);
      end for;
      f := z*f;
  else
      for i:=1 to #as0 do
	  j := (i-1)*n0;
	  f := f + as0[i]*x^(j)*z^(d-j);
      end for;
      f := x*z*f;
  end if;

  return f,as0;

end function;

function FunctionFromPartialDescent(as0,C,K,k);
  //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli.
  //Output :  The corresponding function defining a divisor.

  S := PolynomialRing(K,2); x := S.1; z := S.2;

  g := S ! 0;
  d := #as0 - 1;
  for i:=1 to #as0 do
      j := (i-1);
      g := g + as0[i]*x^(j)*z^(d-j);
  end for;

  return g;

end function;

function AutomorphismFromPartialDescent(as0,C,K,k);
  //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli.
  //Output :  The corresponding automorphism, as the upper right hand corner of a matrix.

  G,Aut,fG := AutomorphismGroup(K,k);
  for g in G do
      if not g eq Id(G) then
	  sigma := g;
      end if;
  end for;
  sigma := fG(sigma);

  m := #as0;
  r := (sigma(as0[m-1])*as0[1]) / (sigma(as0[m])*as0[2]);
  
  return (k ! r);

end function;

function HyperellipticDescendedDivisorFromDivAndAut(as0,C,g,r,K,k,w,P);
  //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli, along with the data from the previous two functions and a generating element from the quadratic extension incurred and a point on the norm conic.
  //Output :  (Part of) the descended ramification divisor on the canonical model of the quotient of the curve by its automorphisms.

  G,Aut,fG := AutomorphismGroup(K,k);
  for x in G do
      if not x eq Id(G) then
	  sigma := x;
      end if;
  end for;
  sigma := fG(sigma);
  
  //Creating appropriate descent matrix:
  lambda := (P[1]/P[3]) + (P[2]/P[3])*w;
  N := Matrix(K,2,2,[1,sigma(lambda),w,sigma(lambda*w)]);

  //Transforming:
  g0 := TransformForm(N,g);
  g0 := g0 / (Coefficients(g0)[1]);

  return g0;

end function;

function NTo1CoverFromDivAndAut(as0,C,g,r,K,k,w,P);
  //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli, along with the data from the previous two functions and a generating element from the quadratic extension incurred and a point on the norm conic.
  //Output :  The cyclic cover whose pullback furnishes a model for the hyperelliptic quotient of the curve.

  i0,j0,m0,n0 := Explode(C);
  S := Parent(g); x := S.1; z := S.2;

  //Creation of the polynomials:
  pq := (x + (1/w)*z)^n0;
  p := S ! 0;
  q := S ! 0;
  for i:=0 to n0 do
      if IsEven(i) then
	  p := p + MonomialCoefficient(pq, x^i*z^(n0-i))*x^i*z^(n0-i);
      else
	  q := q + MonomialCoefficient(pq, x^i*z^(n0-i))*x^i*z^(n0-i);
      end if;
  end for;
  q := w*q;

  return p,q;

end function;

function HyperellipticDescentFromDivAndAut(as0,C,g,r,K,k);
    //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli, along with the data from the previous two functions.
    //Output :  A boolean indicating whether hyperelliptic descent is possible, and a descended form, if available.

    i0,j0,m0,n0 := Explode(C);
    d0 := m0*n0 + i0;

    //Finding a generator of  K | k  :
    i := 1;
    stop := false;
    while not stop do
	x := K.i;
	if not x in k then
	    stop := true;
	end if;
	i := i + 1;
    end while;
    //An element of trace  0  :
    G,Aut,fG := AutomorphismGroup(K,k);
    for y in G do
	if not y eq Id(G) then
	    sigma := y;
	end if;
    end for;
    sigma := fG(sigma);
    w := (x - sigma(x))/2;
    d := k ! (w^2);

    //Norm equation:
    CNorm := Conic([1, -d, -r]);
    test,P := HasRationalPoint(CNorm);
    
    //Creation of descended form:
    if test then
	S := Parent(g); x := S.1; z := S.2;
	//Descent of the divisor on the base curve  X / Aut (X)  :
	g0 := HyperellipticDescendedDivisorFromDivAndAut(as0,C,g,r,K,k,w,P);
	//Now take an  n0 : 1  cover ramifying over the transformation of the divisor  x*z  .
	p,q := NTo1CoverFromDivAndAut(as0,C,g,r,K,k,w,P);
	h := hom<S -> S | [p,q]>;
	//Pulling pack the divisor:
	g0 := h(g0);
	//If necessary, throw in additional ramification divisor:
	if i0 eq 2 then
	    g0 := g0*(d*x^2 - z^2);
	end if;
	//Coercion to the base field:
	S0 := PolynomialRing(k,2); x := S0.1; z := S0.2;
	f0 := (S0 ! g0);
	f0 := f0 / Coefficients(f0)[1];
	return test,f0;
    end if;

    S0 := PolynomialRing(k,2); x := S0.1; z := S0.2;
    return test, S0 ! 0;

end function;

function NonHyperellipticDescendedDivisorFromDivAndAut(as0,C,g,r,K,k,w);
  //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli, along with the data from the previous two functions and a generating element from the quadratic extension incurred.
  //Output :  (Part of) the descended ramification divisor on the canonical model of the quotient of the curve by its hyperelliptic involution.

  i0,j0,m0,n0 := Explode(C);
  lambda := -1/r;
  mu := k ! (-w^2);

  S2 := Parent(g);
  K2 := FieldOfFractions(S2);
  P1 := Curve(ProjectiveSpace(K,1)); x := P1.1; z := P1.2;
  FP1 := FunctionField(P1); t := FP1.1;
  h := hom<K2 -> FP1 | [t,1]>;
  D := Divisor(P1,h(g));

  //Parametrizing the conic over an extension:
  P2 := ProjectiveSpace(K,2); 
  S3 := CoordinateRing(P2); X := S3.1; Y := S3.2; Z := S3.3;
  Eq0 := X^2 + (1/lambda)*Z^2 + mu*Y^2;
  Eq1 := X^2 + lambda*Y^2 + mu*Z^2;
  Con0 := Curve(P2,Eq0);
  Con1 := Curve(P2,Eq1);
  Param := map<P1 -> Con1 | [w*(-z^2 + lambda*x^2), -2*w*x*z, z^2 + lambda*x^2]>;
  ParamInv := map<Con1 -> P1 | [Y, X - w*Z]>;
  Iso := map<Con0 -> Con1 | [X,(-1/lambda)*Z,Y]>;
  Cov := map<P1 -> P1 | [x^n0 , ((-1/lambda))^((n0 - 1) div 2)*z^n0]>;
  
  //Pulling back divisors and normalizing.
  D := Pullback(Cov,D);
  D := Pullback(ParamInv,D);
  test,f0 := IsPrincipal(D);
  if i0 eq 2 then
      f0 := Y*f0;
  end if;
  D := Pullback(Iso,Divisor(Con1,f0));

  D0,E0 := SignDecomposition(D);
  D0 := Decomposition(D0);

  return Eq0,D0;

end function;

function NonHyperellipticDescentFromDivAndAut(as0,C,g,r,K,k);
  //Input :   A partial descent, its case distinction, its field of definition, and its field of moduli, along with the data from the previous two functions.
  //Output :  A non-hyperelliptic descent, if available.

  i0,j0,m0,n0 := Explode(C);
  lambda := -1/r;

  //Finding a generator of  K | k   and using it to determine mu:
  i := 1;
  stop := false;
  while not stop do
      x := K.i;
      if not x in k then
	  stop := true;
      end if;
      i := i + 1;
  end while;
  //An element of trace  0  :
  G,Aut,fG := AutomorphismGroup(K,k);
  for Gelt in G do
      if not Gelt eq Id(G) then
	  sigma := Gelt;
      end if;
  end for;
  sigma := fG(sigma);
  w := (x - sigma(x))/2;
  mu := k ! (-w^2);

  return NonHyperellipticDescendedDivisorFromDivAndAut(as0,C,g,r,K,k,w);

end function;

function Descent(f);
  //Input :   A ternary quartic  f  .
  //Output :  If it exists, a hyperelliptic descent over the field of moduli; if that does not exists, then a general descent, if it exists; if neither exists, then a model over a quadratic extension of the field of moduli. The field of definition of the arithmetic diagonal invariants and the field of moduli are also returned.

  test,f3,alpha := HasCyclicAutomorphismGroup(f);
  if not test then
      "A descent exists and is available by other methods.";
      return f, BaseField(Parent(f)),BaseField(Parent(f));
  end if;
  
  f2 := StandardForm(f3,alpha);
  C := CaseDistinction(f2,2);
  as := FundamentalCoefficients(f2,C);
  i0,j0,m0,n0 := Explode(C);
  if i0 eq 1 then
      f2 := NormalizeCaseI1(f2);
      I,WI,J,WJ,K,k := GenericNormalizedADI(as,C);
      "A hyperelliptic descent to the field of moduli exists and is given by:";
      f1,as1 := GenericPartialDescent(as,C,J,WJ,K);
      f1;
      k;
      return f1,K,k;
  end if;
  
  I,WI,J,WJ,K,k := GenericNormalizedADI(as,C);
  if K eq k then
      "A hyperelliptic descent to the field of moduli exists and is given by:";
      f1,as1 := GenericPartialDescent(as,C,J,WJ,K);
      return f1,K,k;
  end if;

  f1,as1 := GenericPartialDescent(as,C,J,WJ,K);
  g := FunctionFromPartialDescent(as1,C,K,k);
  r := AutomorphismFromPartialDescent(as1,C,K,k);
  test,f0 := HyperellipticDescentFromDivAndAut(as1,C,g,r,K,k);
  if test then
      "A hyperelliptic descent to the field of moduli exists and is given by:";
      f0;
      k;
      return f0,K,k;
  end if;

  d := Degree(f);
  g := (d div 2) - 1;
  if IsOdd(g) and IsOdd(n0) then
      "A hyperelliptic descent to the field of moduli does not exist. A non-hyperelliptic descent to this field does exist and is given by ramifying as follows:";
      Eq0,D0 := NonHyperellipticDescentFromDivAndAut(as1,C,g,r,K,k);
      Eq0;
      D0;
      k;
      return [Eq0,D0],K,k;
  end if;

  "No descent to the field of moduli exists. A partial descent to a quadratic extension of the field of moduli is given by:";
  f1;
  K;
  return f1,K,k;

end function;
